@@ -2,6 +2,7 @@
 from __future__ import print_function, division
 
 import ast
+import contextlib
 import dis
 import inspect
 import json
@@ -21,7 +22,10 @@ from tests.utils import tester, subscript_item, in_finally
 PYPY = 'pypy' in sys.version.lower()
 
 from executing import Source, only, NotOneValueFound
-from executing.executing import PY3, get_instructions
+from executing.executing import PY3, get_instructions, function_node_types
+
+if eval("0"):
+    global_never_defined = 1
 
 
 class TestStuff(unittest.TestCase):
@@ -40,25 +44,74 @@ class TestStuff(unittest.TestCase):
         # @formatter:on
 
     def test_decorator(self):
-        @empty_decorator
-        @decorator_with_args(tester('123'), x=int())
-        @tester(list(tuple([1, 2])), returns=empty_decorator)
-        @tester(
+        @empty_decorator  # 0
+        @decorator_with_args(tester('123'), x=int())  # 1
+        @tester(list(tuple([1, 2])))  # 2!
+        @tester(  # 3!
             list(
                 tuple(
                     [3, 4])),
-            returns=empty_decorator)
-        @empty_decorator
-        @decorator_with_args(
+            )
+        @empty_decorator  # 4
+        @decorator_with_args(  # 5
             str(),
             x=int())
-        @tester(list(tuple([5, 6])), returns=empty_decorator)
-        @tester(list(tuple([7, 8])), returns=empty_decorator)
+        @tester(list(tuple([5, 6])))  # 6!
+        @tester(list(tuple([7, 8])))  # 7!
         @empty_decorator
         @decorator_with_args(tester('sdf'), x=tester('123234'))
         def foo():
             pass
 
+        tester.check_decorators([7, 6, 3, 2])
+
+        empty_decorator.tester = tester
+
+        @empty_decorator
+        @tester
+        @empty_decorator
+        @tester.qwe
+        @empty_decorator
+        @tester("1")
+        @empty_decorator.tester("2")
+        @empty_decorator
+        def foo2(_=tester("3"), __=tester("4")):
+            pass
+
+        tester.check_decorators([6, 5, 3, 1])
+
+        @tester
+        @empty_decorator
+        @tester.qwe
+        @empty_decorator
+        @tester("11")
+        @empty_decorator.tester("22")
+        @empty_decorator
+        class foo3(tester("5") and list):
+            pass
+
+        tester.check_decorators([5, 4, 2, 0])
+
+        class Foo(object):
+            @tester
+            @tester
+            @empty_decorator
+            @tester.qwe
+            @empty_decorator
+            def foo(self):
+                super(Foo, self)
+
+                class Bar:
+                    @tester
+                    @empty_decorator
+                    @tester.qwe
+                    @empty_decorator
+                    def bar(self):
+                        pass
+
+        Foo().foo()
+        tester.check_decorators([3, 1, 0, 2, 0])
+
     def test_comprehensions(self):
         # Comprehensions can be separated if they contain different names
         str([{tester(x) for x in [1]}, {tester(y) for y in [1]}])
@@ -313,6 +366,38 @@ class TestStuff(unittest.TestCase):
         check(3)
         check(5)
 
+    @contextlib.contextmanager
+    def assert_name_error(self):
+        try:
+            yield
+        except NameError as e:
+            tb = sys.exc_info()[2]
+            ex = Source.executing(tb.tb_next)
+            self.assertEqual(type(ex.node), ast.Name)
+            self.assertIn(ex.node.id, str(e))
+            self.assertEqual(ex.text(), ex.node.id)
+        else:
+            self.fail("NameError not raised")
+
+    def test_names(self):
+        with self.assert_name_error():
+            self, completely_nonexistent  # noqa
+
+        with self.assert_name_error():
+            self, global_never_defined  # noqa
+
+        with self.assert_name_error():
+            self, local_not_defined_yet  # noqa
+
+        local_not_defined_yet = 1  # noqa
+
+        def foo():
+            with self.assert_name_error():
+                self, closure_not_defined_yet  # noqa
+
+        foo()
+        closure_not_defined_yet = 1  # noqa
+
 
 def is_unary_not(node):
     return isinstance(node, ast.UnaryOp) and isinstance(node.op, ast.Not)
@@ -327,21 +412,25 @@ class TimeOut(Exception):
     'These tests are very slow, enable them explicitly',
 )
 class TestFiles(unittest.TestCase):
+
+    maxDiff = None
+
     def test_files(self):
         self.start_time = time.time()
         root_dir = os.path.dirname(__file__)
         samples_dir = os.path.join(root_dir, 'samples')
-        result_filename = PYPY * 'pypy' + sys.version[:3] + '.json'
+        result_filename = PYPY * 'pypy' + '.'.join(map(str, sys.version_info[:2])) + '.json'
         result_filename = os.path.join(root_dir, 'sample_results', result_filename)
         result = {}
 
         for filename in os.listdir(samples_dir):
             full_filename = os.path.join(samples_dir, filename)
-            result[filename] = self.check_filename(full_filename)
+            result[filename] = self.check_filename(full_filename, check_names=True)
 
         if os.getenv('FIX_EXECUTING_TESTS'):
             with open(result_filename, 'w') as outfile:
                 json.dump(result, outfile, indent=4, sort_keys=True)
+            return
         else:
             with open(result_filename, 'r') as infile:
                 self.assertEqual(result, json.load(infile))
@@ -370,11 +459,11 @@ class TestFiles(unittest.TestCase):
                 continue
 
             try:
-                self.check_filename(filename)
+                self.check_filename(filename, check_names=False)
             except TimeOut:
                 print("Time's up")
 
-    def check_filename(self, filename):
+    def check_filename(self, filename, check_names):
         print(filename)
         source = Source.for_filename(filename)
 
@@ -386,8 +475,11 @@ class TestFiles(unittest.TestCase):
                     self.assertEqual(code_qualname, qualname)
 
         nodes = defaultdict(list)
+        decorators = defaultdict(list)
+        expected_decorators = {}
         for node in ast.walk(source.tree):
             if isinstance(node, (
+                    (ast.Name,) * check_names,
                     ast.UnaryOp,
                     ast.BinOp,
                     ast.Subscript,
@@ -396,20 +488,25 @@ class TestFiles(unittest.TestCase):
                     ast.Attribute
             )):
                 nodes[node] = []
+            elif isinstance(node, (ast.ClassDef, function_node_types)):
+                expected_decorators[(node.lineno, node.name)] = node.decorator_list[::-1]
+                decorators[(node.lineno, node.name)] = []
 
         code = compile(source.tree, source.filename, 'exec')
-        result = list(self.check_code(code, nodes))
+        result = list(self.check_code(code, nodes, decorators, check_names=check_names))
 
         if not re.search(r'^\s*if 0(:| and )', source.text, re.MULTILINE):
             for node, values in nodes.items():
                 if is_unary_not(node):
                     continue
 
-                if isinstance(getattr(node, 'ctx', None), (ast.Store, ast.Del)):
+                if isinstance(getattr(node, 'ctx', None), (ast.Store, ast.Del, getattr(ast, 'Param', ()))):
                     assert not values
                     continue
 
                 if isinstance(node, ast.Compare):
+                    if sys.version_info >= (3, 10):
+                        continue
                     if len(node.ops) > 1:
                         assert not values
                         continue
@@ -420,7 +517,9 @@ class TestFiles(unittest.TestCase):
                 if is_literal(node):
                     continue
 
-                if sys.version_info >= (3, 9) and in_finally(node):
+                if sys.version_info >= (3, 10):
+                    correct = len(values) >= 1
+                elif sys.version_info >= (3, 9) and in_finally(node):
                     correct = len(values) > 1
                 else:
                     correct = len(values) == 1
@@ -432,7 +531,7 @@ class TestFiles(unittest.TestCase):
 
         return result
 
-    def check_code(self, code, nodes):
+    def check_code(self, code, nodes, decorators, check_names):
         linestarts = dict(dis.findlinestarts(code))
         instructions = get_instructions(code)
         lineno = None
@@ -442,10 +541,13 @@ class TestFiles(unittest.TestCase):
                 raise TimeOut
 
             lineno = linestarts.get(inst.offset, lineno)
-            if not inst.opname.startswith((
+            if not inst.opname.startswith(
+                (
                     'BINARY_', 'UNARY_', 'LOAD_ATTR', 'LOAD_METHOD', 'LOOKUP_METHOD',
                     'SLICE+', 'COMPARE_OP', 'CALL_', 'IS_OP', 'CONTAINS_OP',
-            )):
+                )
+                + ('LOAD_NAME', 'LOAD_GLOBAL', 'LOAD_FAST', 'LOAD_DEREF', 'LOAD_CLASSDEREF') * check_names
+            ):
                 continue
             frame = C()
             frame.f_lasti = inst.offset
@@ -457,24 +559,40 @@ class TestFiles(unittest.TestCase):
 
             try:
                 try:
-                    node = Source.executing(frame).node
+                    ex = Source.executing(frame)
+                    node = ex.node
                 except Exception:
-                    if inst.opname.startswith(('COMPARE_OP', 'CALL_')):
+                    if inst.opname.startswith(('COMPARE_OP', 'IS_OP', 'CALL_', 'LOAD_NAME')):
                         continue
-                    if isinstance(only(source.statements_at_line(lineno)), (ast.AugAssign, ast.Import)):
+                    if inst.opname == 'LOAD_FAST' and inst.argval == '.0':
+                        continue
+
+                    if inst.argval == 'AssertionError':
+                        continue
+
+                    if any(
+                        isinstance(stmt, (ast.AugAssign, ast.Import))
+                        for stmt in source.statements_at_line(lineno)
+                    ):
                         continue
                     raise
+                # argval isn't set for all relevant instructions in python 2
+                if isinstance(node, ast.Name) and (PY3 or inst.argval):
+                    self.assertEqual(inst.argval, node.id)
             except Exception:
                 print(source.text, lineno, inst, node and ast.dump(node), code, file=sys.stderr, sep='\n')
                 raise
 
-            nodes[node].append((inst, frame.__dict__))
+            if ex.decorator:
+                decorators[(node.lineno, node.name)].append(ex.decorator)
+            else:
+                nodes[node].append((inst, frame.__dict__))
 
-            yield [inst.opname, node_string(source, node)]
+            yield [inst.opname, node_string(source, ex.decorator or node)]
 
         for const in code.co_consts:
             if isinstance(const, type(code)):
-                for x in self.check_code(const, nodes):
+                for x in self.check_code(const, nodes, decorators, check_names=check_names):
                     yield x
 
 
